package info.textgrid.middleware;

import java.io.IOException;
import java.io.StringWriter;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.XMLGregorianCalendar;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.IdentifyType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.ListMetadataFormatsType;
import info.textgrid.middleware.oaipmh.ListRecordsType;
import info.textgrid.middleware.oaipmh.ListSetsType;
import info.textgrid.middleware.oaipmh.OAIPMHtype;
import info.textgrid.middleware.oaipmh.ObjectFactory;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.VerbType;

/**
 * <p>
 * This class takes all possible arguments of an OAI-PMH request and proceeds depended on the given
 * arguments. It also deals as response handler.
 * </p>
 * 
 * @author Maximilian Brodhun, SUB Göttingen
 * @author Stefan E. Funk, SUB Göttingen
 * @version 2021-10-27
 * @since 2014-01-29
 */
public class OAIPMHImpl implements OAIPMHProducer {

  // NOTE Changed OpenAireIdentifierList into IdentifierListDelivererDATACITE!

  // **
  // PRIVATE STATIC FINAL
  // **

  private static final String ERR_VAL_IDENTIFIER = "identifier";
  private static final String ERR_VAL_FROM = "from";
  private static final String ERR_VAL_MDPREFIX = "metadataPrefix";
  private static final String ERR_VAL_RESTOK = "resumptionToken";
  private static final String ERR_VAL_UNTIL = "until";
  private static final String ERR_VAL_SET = "set";
  private static final String ILLEGAL_OAI_VERB = "Illegal OAI verb";
  private static final String ILLEGAL_ARG = "The request includes illegal arguments: ";

  // **
  // STATICS
  // **

  private static ObjectFactory obf = new ObjectFactory();

  // **
  // PRIVATES
  // **

  private ErrorHandler error = new ErrorHandler();
  private Log log = LogFactory.getLog(OAIPMHImpl.class);
  private RepIdentification rep;

  private RecordListDelivererDC recordListDC;
  private RecordListDelivererIDIOM recordListIDIOM;
  private RecordListDelivererDATACITE recordListDATACITE;

  private RecordDelivererDC recordDC;
  private RecordDelivererIDIOM recordIDIOM;
  private RecordDelivererDATACITE recordDATACITE;

  private IdentifierListDelivererDC identifierListDC;
  private IdentifierListDelivererIDIOM identifierListIDIOM;
  private IdentifierListDelivererDATACITE identifierListDATACITE;

  private MetadataFormatListDelivererInterface metadataFormatList;
  private SetListDeliverer setList;

  private IDIOMImages imageList = new IDIOMImages();
  // **
  // PUBLIC
  // **

  public boolean textgrid;
  public boolean dariah;

  /**
   * @param rep
   * @param recordDC
   * @param recordIDIOM
   * @param recordDATACITE
   * @param recordListDC
   * @param recordListIDIOM
   * @param recordListDATACITE
   * @param metadataFormatList
   * @param setList
   * @param identifierList
   * @param identifierListIDIOM
   * @param identifierListDATACITE
   */
  public OAIPMHImpl(
      RepIdentification rep, // 0
      RecordDelivererDC recordDC, // 1
      RecordDelivererIDIOM recordIDIOM, // 2
      RecordDelivererDATACITE recordDATACITE, // 3
      RecordListDelivererDC recordListDC, // 4
      RecordListDelivererIDIOM recordListIDIOM, // 5
      RecordListDelivererDATACITE recordListDATACITE, // 6
      MetadataFormatListDelivererInterface metadataFormatList, // 7
      SetListDeliverer setList, // 8
      IdentifierListDelivererDC identifierList, // 9
      IdentifierListDelivererIDIOM identifierListIDIOM, // 10
      IdentifierListDelivererDATACITE identifierListDATACITE, // 11
      IDIOMImages imageList
  ) {

    this.rep = rep;

    this.recordDC = recordDC;
    this.recordIDIOM = recordIDIOM;
    this.recordDATACITE = recordDATACITE;

    this.recordListDC = recordListDC;
    this.recordListIDIOM = recordListIDIOM;
    this.recordListDATACITE = recordListDATACITE;

    this.identifierListDC = identifierList;
    this.identifierListIDIOM = identifierListIDIOM;
    this.identifierListDATACITE = identifierListDATACITE;

    this.metadataFormatList = metadataFormatList;

    this.setList = setList;
    this.imageList = imageList;
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.OAIPMHProducer#postRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public String postRequest(final String verb, final String identifier, final String metadataPrefix,
      final String set, final String from, final String until, final String resumptionToken) {

    String result = "";

    // Check IDs.
    String checkedID = OAIPMHUtilities.omitPrefixFromIdentifier(identifier);
    String checkedSET = OAIPMHUtilities.omitPrefixFromIdentifier(set);

    try {
      result = getStringFromJAXBOAIElement(verb,
          handleRequest(
              verb,
              checkedID,
              metadataPrefix,
              checkedSET,
              from,
              until,
              resumptionToken));

    } catch (ParseException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.OAIPMHProducer#getRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public String getRequest(final String verb, final String identifier, final String metadataPrefix,
      final String set, final String from, final String until, final String resumptionToken) {

    String result = "";

    // Check IDs.
    String checkedID = OAIPMHUtilities.omitPrefixFromIdentifier(identifier);
    String checkedSET = OAIPMHUtilities.omitPrefixFromIdentifier(set);

    try {
      result = getStringFromJAXBOAIElement(verb,
          handleRequest(
              verb,
              checkedID,
              metadataPrefix,
              checkedSET,
              from,
              until,
              resumptionToken));

    } catch (ParseException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

  /**
   * @return
   */
  public String getVersion() {
    return OaipmhServiceVersion.BUILDNAME + "-" + OaipmhServiceVersion.VERSION + "+"
        + OaipmhServiceVersion.BUILDDATE;
  }

  /**
   * <p>
   * Checks the Identify request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype identifyRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    List<String> errorValues = new ArrayList<String>();
    IdentifyType tgRepIdentificationRequest = new IdentifyType();

    if (this.rep.requestChecker(request)) {

      tgRepIdentificationRequest.setBaseURL(this.rep.getBaseUrl());
      tgRepIdentificationRequest.setDeletedRecord(this.rep.getDeletedRecordStatus());
      tgRepIdentificationRequest.setEarliestDatestamp(this.rep.getEarliestDatestamp());
      tgRepIdentificationRequest.setGranularity(this.rep.getGranularity());
      tgRepIdentificationRequest.setRepositoryName(this.rep.getRepositoryName());
      tgRepIdentificationRequest.setProtocolVersion(this.rep.getProtocolVersion());
      tgRepIdentificationRequest.getAdminEmail().add(this.rep.getAdminMaiAddresss());
      oaipmhRoot.setIdentify(tgRepIdentificationRequest);

    } else {
      if (request.getIdentifier() != null) {
        errorValues.add(ERR_VAL_IDENTIFIER);
      }
      if (request.getFrom() != null) {
        errorValues.add(ERR_VAL_FROM);
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add(ERR_VAL_MDPREFIX);
      }
      if (request.getResumptionToken() != null) {
        errorValues.add(ERR_VAL_RESTOK);
      }
      if (request.getUntil() != null) {
        errorValues.add(ERR_VAL_FROM);
      }
      if (request.getSet() != null) {
        errorValues.add(ERR_VAL_SET);
      }

      this.error.setError(OAIPMHConstants.OAI_BAD_ARGUMENT, ILLEGAL_ARG + errorValues);
      oaipmhRoot.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Checks the GetRecord request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype getRecordRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    // Check request. In GetRecord: allowed parameters: identifier, metadataPrefix, verb forbidden:
    // resumptionToken, from, till, set optional: no one.
	  
	
    ErrorHandler requestErrors = RecordDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getCode() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
    }

    // Get Record Request is done HERE!
    else {

      // Default is DC.
      RecordDelivererInterface recDeliv = this.recordDC;

      // Take IDIOM if IDIOM prefix.
      if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_PREFIX)) {
        recDeliv = this.recordIDIOM;
      }
      if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_OPENAIRE_PREFIX)) {
        recDeliv = this.recordDATACITE;
      }
      if(request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_IMAGE_PREFIX)){  
        recDeliv = new IDIOMImages();
      }

      // Finally start the QUERY!
      GetRecordType getRecord = new GetRecordType();

      String id = request.getIdentifier();
      try {
        getRecord = recDeliv.getRecordById(id);
      } catch (ParseException | DatatypeConfigurationException | IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      if (getRecord != null) {
        oaipmhRoot.setGetRecord(getRecord);
      } else {
        ErrorHandler idError = new ErrorHandler();
        idError.setError(OAIPMHConstants.OAI_ID_DOES_NOT_EXIST, "No matching identifier ");
        oaipmhRoot.getError().add(idError.getError());
      }
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Checks the ListIdentifiers request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oai
   * @param request
   * @return
   * @throws ParseException
   */
  public OAIPMHtype listIdentifiersRequest(OAIPMHtype oaipmhRoot, RequestType request)
      throws ParseException {
    IDIOMImages imm = new IDIOMImages();
    // Check request first.
    ErrorHandler requestErrors = IdentifierListDelivererAbstract.requestChecker(request);
    if (requestErrors.getError().getValue() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
      return oaipmhRoot;
    } else {

      IdentifierListDelivererInterface idListDeliv = null;

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)) {
          idListDeliv = this.identifierListDC;
        } else if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_PREFIX)) {
          idListDeliv = this.identifierListIDIOM;
        } else if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_OPENAIRE_PREFIX)) {
          idListDeliv = this.identifierListDATACITE;
        } else if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_IMAGE_PREFIX)) {          
          imm.setKindID("1");
          imm.setPageNumber("1");
          imm.setPerPage("100");
          idListDeliv = imm;
        }
      }

      // If metadata prefix is NOT set, check resumption token hash maps to decide which metadata
      // format we shall use.
      else {
        // Token is from DC request.
        if (IdentifierListDelivererDC.cursorCollector != null
            && IdentifierListDelivererDC.cursorCollector
                .containsKey(request.getResumptionToken())) {
          idListDeliv = this.identifierListDC;
        }
        // Token is from IDIOM request.
        else if (IdentifierListDelivererIDIOM.cursorCollector != null
            && IdentifierListDelivererIDIOM.cursorCollector
                .containsKey(request.getResumptionToken())) {
          idListDeliv = this.identifierListIDIOM;        }
        // Token is from DATACITE request.
        else if (IdentifierListDelivererDATACITE.cursorCollector != null
            && IdentifierListDelivererDATACITE.cursorCollector
                .containsKey(request.getResumptionToken())) {
          idListDeliv = this.identifierListDATACITE;
        }
         // Token is from IDIOMImages request.
       else if (IDIOMImages.cursorCollector != null
          && IDIOMImages.cursorCollector
                 .containsKey(request.getResumptionToken())) {
              idListDeliv = imm;
              imm.setKindID("1");
              imm.setPerPage("100");
              imm.setPageNumber(Integer.toString(IDIOMImages.cursorCollector.get(request.getResumptionToken())));           
            }
        // We have got an invalid resumptionToken here!
        else {
          ErrorHandler e = new ErrorHandler();
          e.setError(OAIPMHConstants.OAI_BAD_RESUMPTION_TOKEN,
              "The value of the resumptionToken is invalid or expired.");
          oaipmhRoot.getError().add(e.getError());
          return oaipmhRoot;
        }
      }

      ListIdentifiersType listIdentifiers = null;
      try {
        listIdentifiers = idListDeliv.processIdentifierList(request.getFrom(), request.getUntil(),
            request.getSet(), request.getResumptionToken());
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      if (listIdentifiers != null) {
        oaipmhRoot.setListIdentifiers(listIdentifiers);
      } else {
        ErrorHandler idError = new ErrorHandler();
        idError.setError(OAIPMHConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oaipmhRoot.getError().add(idError.getError());
      }
    }

    return oaipmhRoot;

  }

  /**
   * <p>
   * Checks the ListSets request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oai
   * @param request
   * @return
   */
  public OAIPMHtype listSetsRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    ListSetsType sl = this.setList.setListBuilder();

    if (this.setList.requestChecker(request)) {
      oaipmhRoot.setListSets(sl);
    } else {
      List<String> errorValues = new ArrayList<String>();
      ErrorHandler badArgument = new ErrorHandler();
      if (request.getIdentifier() != null) {
        errorValues.add(ERR_VAL_IDENTIFIER);
      }
      if (request.getFrom() != null) {
        errorValues.add(ERR_VAL_FROM);
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add(ERR_VAL_MDPREFIX);
      }
      if (request.getResumptionToken() != null) {
        errorValues.add(ERR_VAL_RESTOK);
      }
      if (request.getUntil() != null) {
        errorValues.add(ERR_VAL_UNTIL);
      }
      if (request.getSet() != null) {
        errorValues.add(ERR_VAL_SET);
      }

      badArgument.setError(OAIPMHConstants.OAI_BAD_ARGUMENT, ILLEGAL_ARG + errorValues);
      oaipmhRoot.getError().add(badArgument.getError());
    }

    return oaipmhRoot;
  }

  /**
   * @param id
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype listMetadataFormatsRequest(String id, OAIPMHtype oaipmhRoot,
      RequestType request) {

    List<String> errorValues = new ArrayList<String>();
    ListMetadataFormatsType listMF = new ListMetadataFormatsType();

    if (this.metadataFormatList.requestChecker(request)) {

      if (id.isEmpty()) {
        listMF = this.metadataFormatList.setMetadataFormatList();
      } else {
        listMF = this.metadataFormatList.setMetadataFormatList(request.getIdentifier());
      }

      if (listMF != null) {
        oaipmhRoot.setListMetadataFormats(listMF);
      } else {
        this.error.setError(OAIPMHConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oaipmhRoot.getError().add(this.error.getError());
      }
    }

    else {
      if (request.getFrom() != null) {
        errorValues.add(ERR_VAL_FROM);
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add(ERR_VAL_MDPREFIX);
      }
      if (request.getResumptionToken() != null) {
        errorValues.add(ERR_VAL_RESTOK);
      }
      if (request.getSet() != null) {
        errorValues.add(ERR_VAL_SET);
      }
      if (request.getUntil() != null) {
        errorValues.add(ERR_VAL_UNTIL);
      }

      this.error.setError(OAIPMHConstants.OAI_BAD_ARGUMENT, "The request includes illegal "
          + "arguments or is missing required arguments: " + errorValues);
      oaipmhRoot.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * @param oai
   * @param request
   * @return
   * @throws ParseException
   * @throws IOException
   */
  public OAIPMHtype listRecordsRequest(OAIPMHtype oaipmhRoot, RequestType request)
      throws ParseException, IOException {
    IDIOMImages imm = new IDIOMImages();
    // Check request first.
    ErrorHandler requestErrors = RecordListDelivererAbstract.requestChecker(request);
    if (requestErrors.getError().getCode() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
      return oaipmhRoot;
    } else {

      RecordListDelivererInterface recListDeliv = null;

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)) {
          recListDeliv = this.recordListDC;
        }
        if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_PREFIX)) {
          recListDeliv = this.recordListIDIOM;
        }
        if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_OPENAIRE_PREFIX)) {
          recListDeliv = this.recordListDATACITE;
        }
        if (request.getMetadataPrefix().equals(OAIPMHConstants.METADATA_IDIOM_IMAGE_PREFIX)) {
          imm.setKindID("1");
          imm.setPageNumber("1");
          imm.setPerPage("100");
          recListDeliv = imm;
        }
      }

      // If metadata prefix is NOT set, check resumption token hash maps to decide which metadata
      // format we shall use.
      else {

        // Token is from DC request.
        if (RecordListDelivererDC.cursorCollector != null
            && RecordListDelivererDC.cursorCollector.containsKey(request.getResumptionToken())) {
          recListDeliv = this.recordListDC;
        }
        // Token is from IDIOM request.
        else if (RecordListDelivererIDIOM.cursorCollector != null
            && RecordListDelivererIDIOM.cursorCollector.containsKey(request.getResumptionToken())) {
          recListDeliv = this.recordListIDIOM;
        }
        // Token is from DATACITE request.
        else if (RecordListDelivererDATACITE.cursorCollector != null
            && RecordListDelivererDATACITE.cursorCollector
                .containsKey(request.getResumptionToken())) {
          recListDeliv = this.recordListDATACITE;
        }
        // Token is from DATACITE request.
        else if (IDIOMImages.cursorCollector != null
            && IDIOMImages.cursorCollector
                .containsKey(request.getResumptionToken())) {
                  imm.setKindID("1");
                  imm.setPerPage("100");
                  imm.setPageNumber(Integer.toString(IDIOMImages.cursorCollector.get(request.getResumptionToken())));
          recListDeliv = imm;
        }
        // We have got an invalid resumptionToken here!
        else {
          ErrorHandler e = new ErrorHandler();
          e.setError(OAIPMHConstants.OAI_BAD_RESUMPTION_TOKEN,
              "The value of the resumptionToken is invalid or expired.");
          oaipmhRoot.getError().add(e.getError());
          return oaipmhRoot;
        }
      }

      ListRecordsType listRecords = recListDeliv.getRecords(
          request.getFrom(),
          request.getUntil(),
          request.getSet(),
          request.getResumptionToken());

      if (listRecords != null) {
        oaipmhRoot.setListRecords(listRecords);
      }
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Taking the string from the REST request and converting into an OAIPMH verb type.
   * </p>
   * 
   * @param verb : what is to within the repository
   * @return verbParam: the verbParam as an OAIPMH verb type
   */
  public VerbType setVerb(String verb) {

    VerbType verbParam = null;

    if (verb != null && verb.equals("Identify")) {
      verbParam = VerbType.IDENTIFY;
    } else if (verb != null && verb.equals("ListMetadataFormats")) {
      verbParam = VerbType.LIST_METADATA_FORMATS;
    } else if (verb != null && verb.equals("ListSets")) {
      verbParam = VerbType.LIST_SETS;
    } else if (verb != null && verb.equals("ListIdentifiers")) {
      verbParam = VerbType.LIST_IDENTIFIERS;
    } else if (verb != null && verb.equals("ListRecords")) {
      verbParam = VerbType.LIST_RECORDS;
    } else if (verb != null && verb.equals("GetRecord")) {
      verbParam = VerbType.GET_RECORD;
    } else if (verb != null && verb.length() > 0) {
      this.error.setError(OAIPMHConstants.OAI_VERB_ERROR, ILLEGAL_OAI_VERB);
    }
    // Verb not set --> please do use verb=Identify!
    // FIXME Is the following appropriate?? Do we miss any error messages that way??
    else {
      verbParam = VerbType.IDENTIFY;
    }

    return verbParam;
  }

  /**
   * <p>
   * Setting the from value for the request.
   * </p>
   * 
   * @param from
   */
  public void setFromRequestValue(String from, RequestType request) {
    if (!from.isEmpty()) {
      request.setFrom(from);
    }
  }

  /**
   * <p>
   * Setting the until value for the request.
   * </p>
   * 
   * @param until
   */
  public void setUntilRequestValue(String until, RequestType request) {
    if (!until.isEmpty()) {
      request.setUntil(until);
    }
  }

  /**
   * <p>
   * Setting the identifier value for the request.
   * </p>
   * 
   * @param identifier
   */
  public void setIdentifierRequestValue(String identifier, RequestType request) {
    if (!identifier.isEmpty()) {
      request.setIdentifier(identifier);
    }
  }

  /**
   * <p>
   * Setting the metadataPrefix value for the request.
   * </p>
   * 
   * @param metadataPrefix
   */
  public void setMetadataPrefixRequestValue(String metadataPrefix, RequestType request) {
    if (!metadataPrefix.isEmpty()) {
      request.setMetadataPrefix(metadataPrefix);
    }
  }

  /**
   * <p>
   * Setting the resumptionToken value for the request.
   * </p>
   * 
   * @param resumptionToken
   */
  public void setResumptionTokenRequestValue(String resumptionToken, RequestType request) {
    if (!resumptionToken.isEmpty()) {
      request.setResumptionToken(resumptionToken);
    }
  }

  /**
   * <p>
   * Setting the set value for the request
   * </p>
   * 
   * @param set
   * @param request
   */
  public void setSetRequestValue(String set, RequestType request) {
    if (!set.isEmpty()) {
      request.setSet(set);
    }
  }

  /**
   * @param textgrid
   */
  public void setTextGrid(boolean textgrid) {
    this.textgrid = textgrid;
  }

  /**
   * @param dariah
   */
  public void setDariah(boolean dariah) {
    this.dariah = dariah;
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * @param verb
   * @param identifier
   * @param metadataPrefix
   * @param set
   * @param from
   * @param until
   * @param resumptionToken
   * @return
   * @throws ParseException
   * @throws IOException
   */
  private JAXBElement<OAIPMHtype> handleRequest(String verb, String identifier,
      String metadataPrefix, String set, String from, String until, String resumptionToken)
      throws ParseException, IOException {

    JAXBElement<OAIPMHtype> result;

    OAIPMHtype response = new OAIPMHtype();
    VerbType verbParam = setVerb(verb);
    RequestType request = new RequestType();

    // Set the responseDate of today.
    try {
      XMLGregorianCalendar convertedDateFormat;
      convertedDateFormat = OAIPMHUtilities
          .convertDateFormat(OAIPMHUtilities.getXMLGregorianCalendarNow().toString());
      response.setResponseDate(convertedDateFormat);
    } catch (ParseException e) {
      this.log.debug("can not parse date format");
    } catch (DatatypeConfigurationException e) {
      this.log.debug("datatype configuration failed");
    }

    // Convert the request values.
    setFromRequestValue(from, request);
    setUntilRequestValue(until, request);
    setIdentifierRequestValue(identifier, request);
    setMetadataPrefixRequestValue(metadataPrefix, request);
    setResumptionTokenRequestValue(resumptionToken, request);
    setSetRequestValue(set, request);
    request.setVerb(verbParam);

    if (this.textgrid == true) {
      request.setValue(TGConstants.REPOSITORY_BASEURL);
    }
    if (this.dariah == true) {
      request.setValue(OAIPMHConstants.REPOSITORY_BASEURL);
    }

    if (verbParam != null) {
      if (verbParam.value().equals("Identify")) {
        response = identifyRequest(response, request);
      } else if (verbParam.value().equals("GetRecord")) {
        response = getRecordRequest(response, request);
      } else if (verbParam.value().equals("ListIdentifiers")) {
        response = listIdentifiersRequest(response, request);
      } else if (verbParam.value().equals("ListSets")) {
        response = listSetsRequest(response, request);
      } else if (verbParam.value().equals("ListMetadataFormats")) {
        response = listMetadataFormatsRequest(identifier, response, request);
      } else if (verbParam.value().equals("ListRecords")) {
        response = listRecordsRequest(response, request);
      }
    } else {
      if (this.error.getError().getValue() != null
          && this.error.getError().getValue().equals(ILLEGAL_OAI_VERB)) {
        this.error.setError(OAIPMHConstants.OAI_VERB_ERROR, ILLEGAL_OAI_VERB);
        response.getError().add(this.error.getError());
      } else {
        this.error.setError(OAIPMHConstants.OAI_VERB_ERROR, ILLEGAL_OAI_VERB);
        response.getError().add(this.error.getError());
      }
    }

    response.setRequest(request);
    result = obf.createOAIPMH(response);

    return result;
  }

  /**
   * <p>
   * Transforms the JaxB element into a string that contains the xsi:schemaLocation attribute. No
   * other way has been working here! I tried EVERYTHING else! Really!
   * </p>
   * 
   * @param theVerb
   * @param theJAXBOAIElement
   * @return
   */
  public static String getStringFromJAXBOAIElement(final String theVerb,
      final JAXBElement<OAIPMHtype> theJAXBOAIElement) {

    String result = "";

    // Always add OAIPMH schema location.
    String schemaLocations =
        OAIPMHUtilities.OAIPMH_NAMESPACE + " " + OAIPMHUtilities.OAIPMH_SCHEMA_LOCATION;

    // Add OAI_DC and DC schema location if needed.
    if (theVerb != null && (theVerb.equals("GetRecord") || theVerb.endsWith("ListRecords"))) {
      schemaLocations +=
          " " + OAIPMHUtilities.OAIDC_NAMESPACE + " " + OAIPMHUtilities.OAIDC_SCHEMA_LOCATION + " "
              + OAIPMHUtilities.DC_NAMESPACE + " " + OAIPMHUtilities.DC_SCHEMA_LOCATION;

      // Add the openaire datacite schemas as well, please.
      schemaLocations +=
          " " + OAIPMHUtilities.OPEN_AIRE_NAMESPACE + " " + OAIPMHUtilities.OAIPMH_SCHEMA_LOCATION
              + " " +
              OAIPMHUtilities.DATACITE_NAMESPACE + " " + OAIPMHUtilities.DATACITE_SCHEMA_LOCATION;
    }

    // PLEASE NOTE We always get oai_dc and dc prefixes defined in the root element due to the
    // bindings.xml file. Seems to be no problem for the time being.

    try {
      // OBACHT! Maybe GEMOKEL! Add schemaLocations to OAI root element here!
      JAXBContext jc = JAXBContext.newInstance(OAIPMHtype.class);

      Marshaller marshaller = jc.createMarshaller();
      // Comment out if one-line-output wanted!
      marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
      marshaller.setProperty(Marshaller.JAXB_SCHEMA_LOCATION, schemaLocations);

      StringWriter sw = new StringWriter();
      marshaller.marshal(theJAXBOAIElement, sw);
      result = sw.toString();
      sw.close();

    } catch (JAXBException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

}
